# Indexes into unclass(g)
igraph_t_idx_n <- 1L
igraph_t_idx_directed <- 2L
igraph_t_idx_from <- 3L
igraph_t_idx_to <- 4L
igraph_t_idx_oi <- 5L
igraph_t_idx_ii <- 6L
igraph_t_idx_os <- 7L
igraph_t_idx_is <- 8L
igraph_t_idx_attr <- 9L
igraph_t_idx_env <- 10L
igraph_t_idx_max <- 11L

# Indexes into unclass(g)[[igraph_t_idx_attr]]
igraph_attr_idx_unknown <- 1L # integer of length 3, for what purpose?
igraph_attr_idx_graph <- 2L
igraph_attr_idx_vertex <- 3L
igraph_attr_idx_edge <- 4L
