list(
  rd_family_title = list(
    adjacency = "Adjacency matrices",
    attributes = "Vertex, edge and graph attributes",
    bipartite = "Bipartite graphs",
    centrality = "Centrality measures",
    coloring = "Graph coloring",
    community = "Community detection",
    components = "Connected components",
    processes = "Processes on graphs",
    cycles = "Graph cycles",
    foreign = "Foreign format readers",
    games = "Random graph models (games)",
    scg = "Spectral Coarse Graining",
    sgm = "Graph matching",
    `constructor modifiers` = "Constructor modifiers (and related functions)"
  )
)
