
#ifndef IGRAPH_EXPORT_H
#define IGRAPH_EXPORT_H

#define IGRAPH_EXPORT
#define IGRAPH_NO_EXPORT
#define IGRAPH_DEPRECATED
#define IGRAPH_DEPRECATED_EXPORT IGRAPH_EXPORT IGRAPH_DEPRECATED
#define IGRAPH_DEPRECATED_NO_EXPORT IGRAPH_NO_EXPORT IGRAPH_DEPRECATED

#endif /* IGRAPH_EXPORT_H */
